/* main.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include "globals.h"
#include "graphics.h"
#include "gui.h"
#include "memory.h"
#include "misc.h"
#include "render.h"
#include "sync.h"
#include "traj.h"
#include "vis5d.h"
#include "work.h"


#ifndef CPU
#  define CPU 1
#endif
#ifndef MBS
#  define MBS 0
#endif
#ifndef TOPOFILE
#  define TOPOFILE "EARTH.TOPO"
#endif
#ifndef MAPFILE
#  define MAPFILE "OUTLSUPW"
#endif


#define MEGA 1024*1024
#define MAX(A,B)  ( (A) > (B) ? (A) : (B) )


void terminate();


#ifdef sgi
#  include <sys/types.h>
#  include <sys/prctl.h>
   int  pid[8];
#endif




/*** main *************************************************************
   Get command line arguments, load data set, call main loop.
**********************************************************************/
main( argc, argv )
int argc;
char *argv[];
{
   char *wdpy_name;
   char topofile[100], mapfile[100], fontname[100];
   int i, box, fontsize, memory, mbs, min_mem;
   int fullflag, alphaflag;
   float ax, ay, az, width;
   struct rlimit rl;

   /* disable core dumps */
/*
   rl.rlim_cur = rl.rlim_max = 0;
   setrlimit( RLIMIT_CORE, &rl );
*/

   if (argc==1) {
      printf("VIS-5D version 3.0\n");
      printf("Usage:\n");
      printf("   vis5d compfile [options]\n");
      printf("Options:\n");
      printf("   -map mapfile\n");
      printf("      Use a different map outline file instead of OUTLSUPW\n");
      printf("      Example:  vis5d complamp -map OUTLUSAL\n");
      printf("   -topo topfile\n");
      printf("      Use a different topography file instead of EARTH.TOPO\n");
      printf("      Example:  vis5d complamp -map MARS.TOPO\n");
      printf("   -box alon alat ahgt\n");
      printf("      Set aspect ratio of the 3-D box (longitude:latitude:height).\n");
      printf("      Example:  vis5d complamp -box 4 2 1\n");
#ifdef sgi
      printf("   -font fontname size\n");
      printf("      Set the font and it's height used in the viewing window.\n");
      printf("      Use the 'listfonts' program to list your system's fonts.\n");
      printf("      Example:  vis5d complamp -font Helvetica 30\n");
#endif
#ifdef stellar
      printf("   -font fontname\n");
      printf("      Set the font used in the viewing window.  Use xlsfonts(1)\n");
      printf("      to list your system's fonts.\n");
      printf("      Example:  vis5d complamp -font fg-30\n");
#endif
#ifdef foo
      /* print the -font option here if you implement it */
#endif
      printf("   -mbs n\n");
      printf("      Limit the memory used by vis5d to 'n' megabytes.  When\n");
      printf("      the limit is exceeded, the least-recently-viewed graphics\n");
      printf("      are deallocated.\n");
#ifdef LEAVEOUT
      printf("   -reverse\n");
      printf("      Display a black 3-D box on a white background instead of\n");
      printf("      white on black.  This can be used with xwd(1) to capture\n");
      printf("      an image for printing.\n");
#endif
      printf("   -rate ms\n");
      printf("      Set the animation rate.  ms is the miniumum delay in\n");
      printf("      milli-seconds between frames.  Default is 100.\n");
      printf("   -full\n");
      printf("      Full-screen window; make the 3-D window as large as possible.\n");
      printf("   -wdpy xdisplay\n");
      printf("      Make widgets appear on a different display.  Useful in\n");
      printf("      combination with -full for making slides and videos.\n");
      printf("      Example:  vis5d COMPLAMP -full -wdpy pluto:0\n");
#ifdef sgi
      printf("   -alpha\n");
      printf("      Use alpha blending instead of screen door transparency.\n");
#endif
      printf("   -wide w\n");
      printf("      Set width of line segments in pixels (default is 1.0).\n");
      exit(0);
   }

   /* set defaults */
   mbs = MBS;
   box = 1;
   AnimRate = ANIMRATE;
   strcpy( topofile, TOPOFILE );
   strcpy( mapfile, MAPFILE );
   wdpy_name = NULL;
   fullflag = 0;
   alphaflag = 0;
   width = 1.0;

   /* look for command line options */
   strcpy( DataFile, argv[1] );

   for (i=2;i<argc;i++) {
      if (strcmp(argv[i],"-box")==0 && i+3<argc) {
	 box = 3;
	 ax = (float) atof( argv[i+1] );
	 ay = (float) atof( argv[i+2] );
	 az = (float) atof( argv[i+3] );
	 i+=3;
      }
      else if (strcmp(argv[i],"-topo")==0 && i+1<argc) {
	 strcpy( topofile, argv[i+1] );
	 i++;
      }
      else if (strcmp(argv[i],"-map")==0 && i+1<argc) {
	 strcpy( mapfile, argv[i+1] );
	 i++;
      }
      else if (strcmp(argv[i],"-d")==0) {
	 Debug = 1;
      }
#ifdef sgi
      else if (strcmp(argv[i],"-font")==0 && i+2<argc) {
         strcpy( fontname, argv[i+1] );
	 fontsize = atoi( argv[i+2] );
	 i+=2;
      }
#endif
#ifdef stellar
      else if (strcmp(argv[i],"-font")==0 && i+1<argc) {
         strcpy( fontname, argv[i+1] );
	 i++;
      }
#endif
#ifdef foo
      /* get the -font option if you implement it */
#endif
      else if (strcmp(argv[i],"-mbs")==0 && i+1<argc) {
	 mbs = atoi( argv[i+1] );
	 i++;
      }
      else if (strcmp(argv[i],"-reverse")==0) {
	 /* swap background and box colors */
	 unsigned long t = BgColor;
	 BgColor = BoxColor;
	 BoxColor = LabelColor = t;
      }
      else if (strcmp(argv[i],"-rate")==0 && i+1<argc) {
	 AnimRate = atoi( argv[i+1] );
	 i++;
      }
      else if (strcmp(argv[i],"-full")==0) {
	 /* make 3-D window as large as possible */
	 fullflag = 1;
      }
#ifdef sgi
      else if (strcmp(argv[i],"-alpha")==0) {
	 /* alpha transparency flag */
	 alphaflag = 1;
      }
#endif
      else if (strcmp(argv[i],"-wdpy")==0 && i+1<argc) {
	 /* widget display */
	 wdpy_name = argv[i+1];
	 i++;
      }
      else if (strcmp(argv[i],"-wide")==0 && i+1<argc) {
	 /* line segment width */
	 width = atof(argv[i+1]);
	 if (width<1.0)
	    width = 1.0;
	 i++;
      }
      else {
	 printf("unknown option: %s\n", argv[i] );
      }
   }

   if (AnimRate<1)
      AnimRate = 1;

   strcpy( ProgramName, "VIS-5D" );

   initialize();
   init_graphics( alphaflag );
   init_sync();
   init_queue();
   ALLOC_LOCK( Gfx_Lock );

   if (mbs==0) {
      init_memory( 0 );  /* use malloc/free */
   }
   else {
      /* Use bounded memory management */
      /* need at least 32 Megabytes: */
/*
      if (mbs<32)
	mbs = 32;
*/
      /* use 75% of system RAM: */
      memory = (int) ( (float) mbs * 0.75 ) * MEGA;
      init_memory( memory );
   }

   if (load_dataset( DataFile )) {

      if (mbs!=0) {
	 /* check if there's enough memory left after loading the data set */
	 min_mem = MAX( memory/3, 5*MEGA );
	 if (mem_available()<min_mem) {
	    printf("Not enough memory (only %d bytes free)\n",mem_available());
	    terminate();
	 }
      }

      initialize2();
      init_traj();
      init_box( box, ax, ay, az );
      TopoFlag = init_topo(topofile);
      MapFlag = init_map(mapfile);
      make_gui( wdpy_name, fontname, fontsize, fullflag );
      set_line_width( width );

      if (Debug)  printf("mem free: %d\n", mem_available() );

/***************** STELLAR: FORK/JOIN *******************/
#ifdef stellar
      /*$dir fork */
      main_loop();
      work(1);
#  if CPU>2
      work(2);
#  endif
#  if CPU>3
      work(3);
#  endif
      /*$dir join */
      ;
#endif

/******************* SGI: SPROC *************************/
#ifdef sgi
      pid[0] = sproc( work, PR_SALL, 1 );
#  if CPU>2
      pid[1] = sproc( work, PR_SALL, 2 );
#  endif
#  if CPU>3
      pid[2] = sproc( work, PR_SALL, 3 );
#  endif
#  if CPU>4
      pid[3] = sproc( work, PR_SALL, 4 );
#  endif
      main_loop();
#endif

/******************* IBM: SINGLE THREADED ***************/
#ifdef ibm
      main_loop();
#endif

/******************* FOO: SINGLE THREADED ***************/
#ifdef foo
      main_loop();
      /* if you can have multiple threads of execution in a process */
      /* change this to something similar to the sgi and stellar code */
#endif

   }
   terminate();

}




/*** initialize *******************************************************
   Initialize global variables BEFORE the dataset has been loaded.
**********************************************************************/
initialize()
{
   int i, j;
   uint c;

   /* init colors */
   for (i=0; i<MAXPARMS; i++) {
      for (j=0; j<5; j++) {
	 if (Color[i][j]==0)
	    Color[i][j] = Color[0][(i*5+j)%21];
      }
   }

   for (i=0;i<MAXPARMS;i++) {
      CloneTable[i] = i;
   }
}




/*** initialize2 ****************************************************
   Initialize graphics-related and other globals vars AFTER the data
   set has been loaded.
**********************************************************************/
initialize2()
{
   int ip, it;

   for (ip=0;ip<MAXPARMS;ip++) {
      for (it=0;it<MAXTIMES;it++) {
         SurfTable[ip][it].valid = 0;
	 HSliceTable[ip][it].valid = 0;
	 VSliceTable[ip][it].valid = 0;
	 CHSliceTable[ip][it].valid = 0;
	 CVSliceTable[ip][it].valid = 0;
      }
   }

   for (ip=0;ip<WINDSLICES;ip++) {
      for (it=0;it<MAXTIMES;it++) {
	 HWindTable[ip][it].valid = 0;
	 VWindTable[ip][it].valid = 0;
      }
   }

   /* Initialize surfaces, slices */
   for (ip=0;ip<NumParms;ip++) {
      IsoLevel[ip] = MinVal[ip];
      HSliceLevel[ip] = (float) (Nl-1) / 2;
      HSliceInterval[ip] = round( (MaxVal[ip] - MinVal[ip]) / 10.0 );
      VSliceR1[ip] = (float) (Nr-1) / 2;
      VSliceC1[ip] = 0.0;
      VSliceR2[ip] = (float) (Nr-1) / 2;
      VSliceC2[ip] = (float) (Nc-1);
      VSliceInterval[ip] = round( (MaxVal[ip] - MinVal[ip]) / 10.0 );
      CHSliceLevel[ip] = (float) (Nl-1) / 2;
      CVSliceR1[ip] = (float) (Nr-1) / 2;
      CVSliceC1[ip] = 0.0;
      CVSliceR2[ip] = (float) (Nr-1) / 2;
      CVSliceC2[ip] = (float) (Nc-1);
   }
   for (ip=0;ip<WINDSLICES;ip++) {
      HWindLevel[ip] = (float) (Nl-1) / 2.0;
      HWindDensity[ip] = HWindScale[ip] = 1.0;
      VWindR1[ip] = (float) (Nr-1) / 2.0;
      VWindC1[ip] = 0.0;
      VWindR2[ip] = (float) (Nr-1) / 2.0;
      VWindC2[ip] = (float) (Nc-1);
      VWindDensity[ip] = VWindScale[ip] = 1.0;
   }

   CursorX = CursorY = CursorZ = 0;
   CursorR = (float) (Nr-1) / 2.0;
   CursorC = (float) (Nc-1) / 2.0;
   CursorL = (float) (Nl-1) / 2.0;
   DisplayCursor = 0;

   DisplayTraj[0] = 1;
}




/*** terminate ********************************************************
   Close sockets, deallocate locks, kill threads and terminate.
**********************************************************************/
void terminate()
{
   FREE_LOCK( Gfx_Lock );
   terminate_queue();
   term_sync();

#ifdef sgi
   if (pid[0])  kill( pid[0], SIGKILL );
#  if CPU>2
      if (pid[1])  kill( pid[1], SIGKILL );
#  endif
#  if CPU>3
      if (pid[2])  kill( pid[2], SIGKILL );
#  endif
#  if CPU>4
      if (pid[3])  kill( pid[3], SIGKILL );
#  endif
#endif
   exit(0);
}





/*** die **************************************************************
   VIS-5D exit point in case of error.
   Input:  msg - a message to print to stdout before exiting.
   Output:  none.
**********************************************************************/
die( msg )
char *msg;
{
   printf("Error: %s\n",msg);
   abort();
}




/*** main_loop ********************************************************
   This is called by main.  It simply updates the 3-D window and
   read user input until the program terminates.
**********************************************************************/
int main_loop()
{
   mat_inv( CTMinv, CTM );
   Redraw = 1;

   while (1) {
      timer(0);  /* reset timer */

      if (Debug) printf("update\n\n");
      while (timer(1) < AnimRate) {
	 get_user_input();
      }

      if (GoTime==1) {
	 CurTime++;
	 if (CurTime==NumTimes)
	    CurTime = 0;
	 Redraw=1;
      }
      else if (GoTime==-1) {
	 CurTime--;
	 if (CurTime<0)
	    CurTime = NumTimes-1;
	 Redraw=1;
      }

      while (Redraw>0) {
	 update_display();
	 Redraw--;
      }

   }

}



/*** yyddd ************************************************************
   Convert a day since Jan 1, 1900 to yyddd format.
**********************************************************************/
int yyddd( iday )
int iday;
{
   int iy, id, iyyddd;

   iy = (4*iday)/1461;
   id = iday-(365*iy+(iy-1)/4);
   iyyddd = iy*1000+id;

   return iyyddd;
}



/*** timer ************************************************************
   Reset the timer or compute frames displayed per second.
   Input:  frames - if 0, reset timer
                    else return frames per second
   Output:  frames per second displayed
**********************************************************************/
int timer(frames)
int       frames;
{
   struct timeval tp;
   static long start_sec, start_usec, lapsed_usec;
   float     msecs;

   /* Get the current time */
   gettimeofday(&tp, (struct timezone *) 0);

   if (frames) {
      /* Compute frames per second */
      lapsed_usec = ((tp.tv_sec - start_sec) * 1000000) +
	tp.tv_usec - start_usec;
      msecs = lapsed_usec / 1000.0;
      return(msecs);
   }
   else {
      /* Reset timer */
      lapsed_usec = 0;
      start_sec = tp.tv_sec;
      start_usec = tp.tv_usec;
      return 0;
   }
}




